# © 2021 Le Filament (<http://www.le-filament.com>)
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).

import itertools

from datetime import datetime

from odoo import _, api, fields, models
from odoo.exceptions import UserError


class Bordereau(models.Model):
    _name = "scop.bordereau"
    _description = "Bordereau des cotisations CG"
    _inherit = "mail.thread"
    _order = "name desc"

    name = fields.Char(
        string="Référence du bordereau",
        readonly=True,
        copy=False,
        default="Brouillon",
        tracking=1,
    )
    version = fields.Integer("Version", default=1, tracking=1)
    base_cotisation_cg = fields.Many2one(
        comodel_name="scop.cotisation.cg",
        string="Base de cotisation",
        required=False,
        readonly=True,
        ondelete="cascade",
    )
    year = fields.Char(
        string="Année du bordereau",
        compute="_compute_year",
        store=True,
        required=False,
    )
    date_cotisation = fields.Date("Date de cotisation")
    is_regul = fields.Boolean("Regularisation", default=False)
    date_regul = fields.Date("Date de régularisation")
    comment_regul = fields.Char("Motif de régularisation")
    partner_id = fields.Many2one(
        comodel_name="res.partner",
        string="Adhérent",
        required=False,
        readonly=True,
        tracking=1,
        index=True,
    )
    member_number = fields.Integer(
        string="N° Adhérent",
        related="partner_id.member_number_int",
        store=True,
        index=True,
    )
    liasse_fiscale_id = fields.Many2one(
        comodel_name="scop.liasse.fiscale",
        string="Liasse fiscale de référence",
        domain="[('partner_id', '=', partner_id), ('is_qualified', '=', True)]",
        tracking=1,
    )
    liasse_count = fields.Integer(
        string="Nombre de liasses", compute="_compute_liasse_count"
    )
    type_liasse_fiscale = fields.Selection(
        string="Type de liasse", related="liasse_fiscale_id.type_id"
    )
    is_liasse_previ = fields.Boolean(
        "Est une liasse prévisionnelle", compute="_compute_is_liasse_previ"
    )
    dureeExercice = fields.Integer(
        string="Durée de l’exercice",
        related="liasse_fiscale_id.dureeExercice",
        tracking=1,
    )

    partner_ur_id = fields.Many2one(
        comodel_name="union.regionale",
        string="UR Adhérent",
        related="partner_id.ur_id",
        store=True,
    )
    payment_mode_id = fields.Many2one(
        comodel_name="account.payment.mode",
        string="Mode de paiment",
        ondelete="restrict",
        readonly=True,
        states={"draft": [("readonly", False)]},
        tracking=1,
    )

    invoice_ids = fields.One2many(
        comodel_name="account.move",
        inverse_name="bordereau_id",
        string="Cotisations",
        required=False,
    )
    invoice_count = fields.Integer(
        string="Nombre d'appels", compute="_compute_invoice_count"
    )

    state = fields.Selection(
        [
            ("new", "Brouillon"),
            ("ongoing", "En cours de modification"),
            ("validated", "Validé"),
            ("paid", "Payé"),
            ("cancel", "Annulé"),
        ],
        string="Statut",
        default="new",
        compute="_compute_state",
        store=True,
        tracking=1,
    )

    company_id = fields.Many2one(
        comodel_name="res.company",
        related="base_cotisation_cg.company_id",
        string="Company",
    )
    company_currency_id = fields.Many2one(
        comodel_name="res.currency",
        related="company_id.currency_id",
        string="Company Currency",
        readonly=True,
    )
    amount_total_cotiz = fields.Monetary(
        string="Montant total de(s) cotisation(s)",
        currency_field="company_currency_id",
        compute="_compute_amount_total_cotiz",
        store=True,
    )
    amount_total_cg = fields.Monetary(
        string="Cotisation CG Scop",
        currency_field="company_currency_id",
        compute="_compute_amount_cg",
        store=True,
    )
    amount_residual = fields.Monetary(
        string="Montant dû",
        currency_field="company_currency_id",
        compute="_compute_amount_residual",
        store=True,
    )
    nb_quarter = fields.Selection(
        string="Nb de trimestres de cotisation",
        selection=[("1", "1"), ("2", "2"), ("3", "3"), ("4", "4")],
        default="4",
        required=True,
        tracking=1,
    )
    details = fields.Html("Détails", compute="_compute_details")

    # Assiettes for cotiz calcul
    year_liasse = fields.Integer(
        string="Année de la liasse",
        related="liasse_fiscale_id.year",
        tracking=1,
    )
    type_assiette = fields.Selection(
        related="liasse_fiscale_id.contribution_base_type", store=True
    )
    montant_assiette = fields.Integer(
        related="liasse_fiscale_id.contribution_base_amount", store=True
    )
    ca = fields.Float(
        string="CA",
        compute="_compute_values_calculation",
        store=True,
        tracking=1,
    )
    va = fields.Float(
        string="VA",
        compute="_compute_values_calculation",
        store=True,
        tracking=1,
    )
    staff_count = fields.Integer(
        string="Effectif de la coop",
        readonly=True,
    )
    staff_shareholder_count = fields.Integer(
        string="Effectif sociétaires de la coop",
        readonly=True,
    )
    staff_average = fields.Float(
        string="Effectif moyen de la coop",
        readonly=True,
    )
    net_results = fields.Float(
        string="Résultat net",
        compute="_compute_values_calculation",
        store=True,
    )
    wage_cg = fields.Float(
        string="Montant masse salariale",
        compute="_compute_values_calculation",
        store=True,
    )
    is_sdd = fields.Boolean(
        "Au prélèvement", compute="_compute_is_sdd", search="_search_is_sdd"
    )
    refund_id = fields.Many2one(
        comodel_name="scop.bordereau",
        string="Nouveau bordereau",
        domain="[('partner_id', '=', partner_id)]",
        tracking=1,
    )
    related_refund_id = fields.Many2one(
        comodel_name="scop.bordereau",
        string="Bordereau initial",
    )
    # Emails management
    recipient_ids = fields.One2many(
        comodel_name="res.partner", compute="_compute_recipient"
    )

    # Historique des versions
    bordereau_version_ids = fields.One2many(
        comodel_name="scop.bordereau.version",
        inverse_name="bordereau_id",
        string="Historique",
    )
    has_outstanding = fields.Boolean(
        string="Paiements en circulation",
        compute="_compute_has_outstanding",
        store=True,
    )
    is_cae = fields.Boolean(
        string="CAE",
        related="partner_id.cae",
        search="_search_is_cae",
    )
    is_federation_com = fields.Boolean(
        string="Fédération de la Communication",
        related="partner_id.is_federation_com",
        search="_search_is_federation_com",
    )
    is_federation_btp = fields.Boolean(
        string="Fédération du BTP",
        related="partner_id.is_federation_btp",
        search="_search_is_federation_btp",
    )
    is_federation_indus = fields.Boolean(
        string="Fédération de l'Industrie",
        related="partner_id.is_federation_indus",
        search="_search_is_federation_indus",
    )

    # ------------------------------------------------------
    # Compute
    # ------------------------------------------------------
    @api.depends("liasse_fiscale_id")
    def _compute_values_calculation(self):
        for r in self:
            if r.liasse_fiscale_id:
                liasse = r.liasse_fiscale_id
                r.ca = liasse.L2052_FL
                r.va = r.base_cotisation_cg.get_va(liasse)
                r.net_results = 0
                if liasse.L2053_HN > 0:
                    r.net_results = liasse.L2053_HN
                elif liasse.L2051_DI > 0:
                    r.net_results = liasse.L2051_DI
                r.wage_cg = liasse.wage_cg
            else:
                r.ca = r.va = r.net_results = r.wage_cg = 0

    @api.depends(
        "invoice_ids",
        "invoice_ids.state",
        "invoice_ids.payment_state",
        "bordereau_version_ids",
        "bordereau_version_ids.state",
    )
    def _compute_state(self):
        for bordereau in self:
            draft_cotiz = bordereau.invoice_ids.filtered(
                lambda cotiz: cotiz.state == "draft"
            )
            ongoing_version = bordereau.bordereau_version_ids.filtered(
                lambda v: v.state == "new"
            )
            if ongoing_version:
                bordereau.state = "ongoing"
            elif draft_cotiz or not bordereau.invoice_ids:
                bordereau.state = "new"
            else:
                unpaid_cotiz = bordereau.invoice_ids.filtered(
                    lambda cotiz: cotiz.payment_state not in ("paid", "cancel")
                )
                if not unpaid_cotiz:
                    bordereau.state = "paid"
                else:
                    bordereau.state = "validated"

    @api.depends("base_cotisation_cg")
    def _compute_year(self):
        for bordereau in self:
            bordereau.year = bordereau.base_cotisation_cg.year

    def _compute_invoice_count(self):
        for r in self:
            r.invoice_count = len(r.invoice_ids)

    def _compute_liasse_count(self):
        for r in self:
            r.liasse_count = len(r.partner_id.liasse_fiscale_ids)

    @api.depends("invoice_ids", "invoice_ids.amount_total_signed")
    def _compute_amount_total_cotiz(self):
        for r in self:
            r.amount_total_cotiz = sum(r.invoice_ids.mapped("amount_total_signed"))

    @api.depends("invoice_ids", "invoice_ids.amount_total_signed")
    def _compute_amount_cg(self):
        type_cotisation_cg = self.env.ref("cgscop_partner.riga_14397")
        for r in self:
            contrib_cg_ids = r.invoice_ids.filtered(
                lambda i: i.type_contribution_id == type_cotisation_cg
            )
            r.amount_total_cg = sum(contrib_cg_ids.mapped("amount_total_signed"))

    @api.depends("invoice_ids", "invoice_ids.amount_residual_signed")
    def _compute_amount_residual(self):
        for r in self:
            r.amount_residual = sum(r.invoice_ids.mapped("amount_residual_signed"))

    def _compute_recipient(self):
        tag_cotiz_id = self.env.company.tag_cotiz_id
        for bordereau in self:
            child_ids = bordereau.partner_id.child_ids.filtered(
                lambda child: (tag_cotiz_id in child.category_id) and child.email
            )
            if bordereau.partner_id.email:
                bordereau.recipient_ids = bordereau.partner_id + child_ids
            else:
                bordereau.recipient_ids = child_ids

    def _compute_type_assiette(self):
        for bordereau in self:
            bordereau.type_assiette = bordereau.base_cotisation_cg.get_type_assiette(
                bordereau.ca, bordereau.va
            )

    def _search_type_assiette(self, operator, value):
        recs = self.search([]).filtered(lambda r: r.type_assiette == value)
        if recs:
            return [("id", "in", [x.id for x in recs])]

    def _compute_is_sdd(self):
        sdd_id = self.env.ref("account_banking_sepa_direct_debit.sepa_direct_debit").id
        for bdx in self:
            if (
                bdx.payment_mode_id
                and bdx.payment_mode_id.payment_method_id.id == sdd_id
            ):
                bdx.is_sdd = True
            else:
                bdx.is_sdd = False

    def _search_is_sdd(self, operator, value):
        recs = self.search([]).filtered(lambda x: x.is_sdd is True)
        if recs:
            return [("id", "in", [x.id for x in recs])]

    def _compute_is_liasse_previ(self):
        for bordereau in self:
            if bordereau.type_liasse_fiscale == "forecast":
                bordereau.is_liasse_previ = True
            else:
                bordereau.is_liasse_previ = False

    @api.depends("invoice_ids.amount_residual")
    def _compute_has_outstanding(self):
        for r in self:
            if r.invoice_ids:
                if r.invoice_ids.filtered(lambda i: i.invoice_has_outstanding is True):
                    r.has_outstanding = True
                else:
                    r.has_outstanding = False
            else:
                r.has_outstanding = False

    def _compute_details(self):
        for r in self:
            contribs = r.invoice_ids.read_group(
                [("id", "in", r.invoice_ids.ids)],
                ["type_contribution_id", "amount_total_signed"],
                ["type_contribution_id"],
            )
            detail = "<table class='o_group o_inner_group'>"
            for contrib in contribs:
                amount_cg = (
                    str("{:,.2f}".format(contrib.get("amount_total_signed")))
                    .replace(",", " ")
                    .replace(".", ",")
                )
                detail += (
                    "<tr><td class='o_td_label font-weight-bold'>{}</td>"
                    "<td style='width: 100%;'>{} €"
                ).format(str(contrib.get("type_contribution_id")[1]), amount_cg)
            detail += "</table><table class='o_group o_inner_group'>"

            for i in range(1, 5):
                # Add amount echeance
                amount_echeance = round(
                    sum(
                        r.invoice_ids.filtered(
                            lambda inv: inv.cotiz_quarter == str(i)
                        ).mapped("amount_total_signed")
                    ),
                    2,
                )
                str_amount = (
                    str("{:,.2f}".format(amount_echeance))
                    .replace(",", " ")
                    .replace(".", ",")
                )
                detail += (
                    "<tr><td class='o_td_label font-weight-bold'>Trimestre {}</td>"
                    "<td style='width: 100%;'>{} €"
                ).format(str(i), str_amount)

                # Check if regul has been done after validation
                type_contribs = r.invoice_ids.mapped("type_contribution_id")
                for type_contrib in type_contribs:
                    is_exo = (
                        len(
                            r.invoice_ids.filtered(
                                lambda inv: inv.cotiz_quarter == i
                                and inv.type_contribution_id == type_contrib
                            )
                        )
                        > 1
                    )
                    if is_exo:
                        detail += " " + '<i class="fa fa-star-o" />'
                        break
                detail += "</td></tr>"
            detail += "</table>"

            r.details = detail

    def _search_is_cae(self, operator, value):
        if operator not in ("=", "!="):
            return []
        else:
            partner_ids = self.env["res.partner"].search([("cae", operator, value)])
            return [("partner_id", "in", partner_ids.ids)]

    def _search_is_federation_com(self, operator, value):
        if operator not in ("=", "!="):
            return []
        else:
            partner_ids = self.env["res.partner"].search(
                [("is_federation_com", operator, value)]
            )
            return [("partner_id", "in", partner_ids.ids)]

    def _search_is_federation_btp(self, operator, value):
        if operator not in ("=", "!="):
            return []
        else:
            partner_ids = self.env["res.partner"].search(
                [("is_federation_btp", operator, value)]
            )
            return [("partner_id", "in", partner_ids.ids)]

    def _search_is_federation_indus(self, operator, value):
        if operator not in ("=", "!="):
            return []
        else:
            partner_ids = self.env["res.partner"].search(
                [("is_federation_indus", operator, value)]
            )
            return [("partner_id", "in", partner_ids.ids)]

    # ------------------------------------------------------
    # Button functions
    # ------------------------------------------------------
    def validate_cotiz_cg(self):
        self.ensure_one()
        quarters = [
            self.base_cotisation_cg.trimester_1,
            self.base_cotisation_cg.trimester_2,
            self.base_cotisation_cg.trimester_3,
            self.base_cotisation_cg.trimester_4,
        ]
        for inv in self.invoice_ids:
            if inv.state == "draft":
                if not inv.invoice_date:
                    inv.update(
                        {
                            "invoice_date": self.date_cotisation,
                        }
                    )
                inv.update(
                    {
                        "payment_mode_id": self.payment_mode_id,
                        "invoice_date_due": quarters[int(inv.cotiz_quarter) - 1],
                    }
                )
                if inv.is_sdd and not inv.mandate_id:
                    raise UserError(
                        _(
                            "Vous ne pouvez pas valider une échéance cotisation au "
                            "prélèvement sans mandat"
                        )
                    )
                else:
                    inv.action_post()

    def button_validate_bordereau(self):
        """
        Appelle la fonction de validation du bordereau
        """
        self.ensure_one()
        return self.validate_bordereau()

    def validate_bordereau(self, mail_compose_message_id=False):
        """
        - Affecte le nom du bordereau avec la séquence
        - Valide les invoices
        - Envoie un mail
        """
        self.ensure_one()
        # Numéro de séquence
        if self.name == "Brouillon":
            self.name = self.env["ir.sequence"].next_by_code("scop.bordereau")
        # Validation des invoices
        self.validate_cotiz_cg()
        # Envoi du mail
        if mail_compose_message_id:
            mail_compose_message = self.env["mail.compose.message"].browse(
                mail_compose_message_id
            )
            mail_compose_message.update(
                {
                    "partner_ids": [(6, 0, self.recipient_ids.ids)],
                    "res_id": self.id,
                }
            )
            mail_compose_message.with_context(mail_notify_force_send=False).send_mail()
        return self.id

    def validate_bordereau_multi(self, mail_compose_message_id):
        """
        Utilise job queue pour valider les bordereaux
        """
        batch_name = (
            fields.Datetime.to_string(fields.Datetime.now()) + " Validation bordereaux "
        )
        batch = self.env["queue.job.batch"].get_new_batch(batch_name)
        for bordereau in self:
            bordereau.with_context(job_batch=batch).with_delay().validate_bordereau(
                mail_compose_message_id
            )
        batch.enqueue()

    def print_bordereau(self):
        for bordereau in self:
            return self.env.ref(
                "cgscop_cotisation_cg.cgscop_bordereau_report"
            ).report_action(bordereau)

    def action_send_email(self):
        self.ensure_one()
        template_id = self.env.ref("cgscop_cotisation_cg.email_template_cotisation_cg")
        ir_model_data = self.env["ir.model.data"]
        try:
            compose_form_id = ir_model_data.get_object_reference(
                "mail", "email_compose_message_wizard_form"
            )[1]
        except ValueError:
            compose_form_id = False
        ctx = {
            "default_model": "scop.bordereau",
            "default_res_id": self.id,
            "default_use_template": True,
            "default_template_id": template_id.id,
        }
        return {
            "type": "ir.actions.act_window",
            "view_type": "form",
            "view_mode": "form",
            "res_model": "mail.compose.message",
            "views": [(compose_form_id, "form")],
            "view_id": compose_form_id,
            "target": "new",
            "context": ctx,
        }

    def update_bordereau_with_liasse(self):
        """
        Supprime ou ajoute les appels de cotisations en fonction du
        nombre de trimestres choisi et de la liasse
        :return:
        """
        self.create_cotiz_and_lines()

    def open_payment(self):
        action_context = {
            "company_ids": [self.company_id.id],
            "partner_ids": [self.partner_id.commercial_partner_id.id],
            "mode": "customers",
        }
        return {
            "type": "ir.actions.client",
            "tag": "manual_reconciliation_view",
            "context": action_context,
        }

    def action_show_cotiz(self):
        """
        :return: Ouvre la vue des cotisations liées au bordereau
        """
        tree_view_id = self.env.ref(
            "cgscop_cotisation.account_move_tree_scop_inherited"
        ).id
        form_view_id = self.env.ref(
            "cgscop_cotisation_cg.invoice_form_scop_cg_inherited"
        ).id
        return {
            "type": "ir.actions.act_window",
            "name": "Appels de cotisations",
            "views": [[tree_view_id, "tree"], [form_view_id, "form"]],
            "view_mode": "form",
            "res_model": "account.move",
            "target": "current",
            "domain": [("id", "in", self.invoice_ids.ids)],
        }

    def action_show_liasse(self):
        """
        :return: Ouvre la vue des liasses liées au partenaire
        """
        return {
            "type": "ir.actions.act_window",
            "name": "Liasses fiscales - " + self.partner_id.name,
            "views": [[False, "tree"], [False, "form"]],
            "view_mode": "form",
            "res_model": "scop.liasse.fiscale",
            "target": "current",
            "domain": [("partner_id", "=", self.partner_id.id)],
        }

    def action_change_liasse(self):
        """
        :return: Ouvre le wizard pour changer la liasse
        """
        wizard = self.env["scop.bordereau.change.liasse.wizard"].create(
            {
                "bordereau_id": self.id,
            }
        )
        return {
            "type": "ir.actions.act_window",
            "name": "Changer la liasse",
            "views": [[False, "form"]],
            "view_mode": "form",
            "res_model": "scop.bordereau.change.liasse.wizard",
            "target": "new",
            "res_id": wizard.id,
        }

    def action_change_payment_mode(self):
        """
        :return: Ouvre le wizard pour changer le mode de paiement
        """
        wizard = self.env["scop.bordereau.change.payment.mode.wizard"].create(
            {
                "bordereau_id": self.id,
            }
        )
        return {
            "type": "ir.actions.act_window",
            "name": "Changer le mode de paiement",
            "views": [[False, "form"]],
            "view_mode": "form",
            "res_model": "scop.bordereau.change.payment.mode.wizard",
            "target": "new",
            "res_id": wizard.id,
        }

    def validate_ongoing_bordereau(self):
        """
        - Valide les invoices en brouillon
        - Met à jour le numéro de version du bordereau
        - Valide le versionnage en cours
        """
        self.ensure_one()
        ongoing_version = self.check_ongoing_version()
        if "-" in self.name:
            new_name = self.name[0:-2] + "-" + str(self.version + 1)
        else:
            new_name = self.name + "-" + str(self.version + 1)
        self.write(
            {
                "name": new_name,
                "version": self.version + 1,
                "date_regul": ongoing_version.date,
                "comment_regul": ongoing_version.comment,
            }
        )
        self.validate_bordereau()
        ongoing_version.write(
            {
                "state": "validated",
            }
        )

    def cancel_ongoing_bordereau(self):
        """
        - Supprime les invoice en brouillon
        - Réaffecte la dernière liasse fiscale (enregistrée dans la version)
        - Supprime le versionnage en cours
        """
        self.ensure_one()
        self.invoice_ids.filtered(lambda i: i.state == "draft").unlink()
        ongoing_version = self.check_ongoing_version()
        self.write({"liasse_fiscale_id": ongoing_version.liasse_fiscale_id_old.id})
        self.invoice_ids.update(
            {"liasse_fiscale_id": ongoing_version.liasse_fiscale_id_old.id}
        )
        ongoing_version.unlink()

    # ------------------------------------------------------
    # Override ORM
    # ------------------------------------------------------
    def unlink(self):
        bordereau_not_new = self.filtered(lambda b: b.state != "new")
        if bordereau_not_new:
            raise UserError(
                _(
                    "Vous ne pouvez pas passer supprimer un bordereau qui a "
                    "déjà été validé.\n Vous pouvez par contre créer des "
                    "avoirs sur les appels de cotisations."
                )
            )
        else:
            return super(Bordereau, self).unlink()

    # ------------------------------------------------------
    # Global Functions
    # ------------------------------------------------------
    def create_cotiz_and_lines(self):
        """
        - Supprime les factures existantes si besoin
        - Crée les appels de cotisation par trimestre
        """
        if self.invoice_ids:
            self.invoice_ids.unlink()

        partner = self.partner_id
        liasse = self.liasse_fiscale_id
        base_cotiz = self.base_cotisation_cg

        # Invoice cotiz CG
        #
        product_cg_id = self.company_id.contribution_cg_id
        type_cotisation_cg = self.env.ref("cgscop_partner.riga_14397").id

        if liasse:
            amount_cg = base_cotiz.round_to_closest_multiple(liasse.contribution_cg, 4)
        else:
            amount_cg = self.env["scop.liasse.fiscale"].get_values_for_cotiz_cg(
                partner
            )["plancher1"]
        self.create_contribution(
            product=product_cg_id,
            type_contribution=type_cotisation_cg,
            amount=amount_cg,
        )

        # Invoice UR et Fédé
        type_cotisation_ur = self.env.ref("cgscop_partner.riga_14399").id

        # Invoice cotiz Fede Com
        #
        if partner.is_federation_com:
            product_fede_com_id = self.company_id.contribution_fede_com_id
            journal_fede_com_id = self.company_id.journal_fede_com_id
            type_cotisation_fede_com = self.env.ref("cgscop_partner.riga_14398").id

            if liasse:
                amount_fede_com = base_cotiz.round_to_closest_multiple(
                    liasse.contribution_com, 4
                )
            else:
                amount_fede_com = self.env["scop.liasse.fiscale"].get_plancher_cotiz()[
                    "fede_com"
                ]
            self.create_contribution(
                product=product_fede_com_id,
                type_contribution=type_cotisation_fede_com,
                amount=amount_fede_com,
                journal_id=journal_fede_com_id,
            )

        # Invoice cotiz Fede CAE
        #
        if partner.cae:
            product_fede_cae_id = self.company_id.contribution_fede_cae_id
            journal_fede_cae_id = self.company_id.journal_fede_cae_id
            type_cotisation_fede_cae = self.env.ref("cgscop_partner.cotiz_fede_cae").id

            if liasse:
                amount_fede_cae = base_cotiz.round_to_closest_multiple(
                    liasse.contribution_cae, 4
                )
            else:
                amount_fede_cae = self.env["scop.liasse.fiscale"].get_plancher_cotiz()[
                    "fede_cae"
                ]
            self.create_contribution(
                product=product_fede_cae_id,
                type_contribution=type_cotisation_fede_cae,
                amount=amount_fede_cae,
                journal_id=journal_fede_cae_id,
            )

        # Invoice cotiz Fede Indus
        #
        if partner.is_federation_indus:
            product_fede_indus_id = self.company_id.contribution_fede_indus_id
            journal_fede_indus_id = self.company_id.journal_fede_indus_id
            type_cotisation_fede_indus = self.env.ref(
                "cgscop_partner.cotiz_fede_indus"
            ).id
            # Calcul de la cotisation de la fédé de l'industrie
            if liasse:
                amount_fede_indus = base_cotiz.round_to_closest_multiple(
                    liasse.contribution_indus, 4
                )
            elif partner.staff_last == 0:
                amount_fede_indus = 4
            else:
                amount_fede_indus = 120

            # Création des factures associées
            self.create_contribution(
                product=product_fede_indus_id,
                type_contribution=type_cotisation_fede_indus,
                amount=amount_fede_indus,
                journal_id=journal_fede_indus_id,
            )

        # Invoice cotiz UR HDF
        #
        ur_hdf = self.env.ref("cgscop_partner.riga_14232").id
        if partner.ur_id.id == ur_hdf:
            product_hdf_id = self.company_id.contribution_hdf_id
            journal_hdf_id = self.company_id.journal_ur_hdf_id

            if liasse:
                amount_hdf = base_cotiz.round_to_closest_multiple(
                    liasse.contribution_hdf, 4
                )
            else:
                amount_hdf = 0
            self.create_contribution(
                product=product_hdf_id,
                type_contribution=type_cotisation_ur,
                amount=amount_hdf,
                journal_id=journal_hdf_id,
            )

        # Invoice cotiz UR Med
        #
        ur_med = self.env.ref("cgscop_partner.riga_14243").id
        if partner.ur_id.id == ur_med:
            product_med_id = self.company_id.contribution_med_id
            journal_ur_med_id = self.company_id.journal_ur_med_id

            if liasse:
                amount_med = base_cotiz.round_to_closest_multiple(
                    liasse.contribution_med, 4
                )
            else:
                amount_med = self.env["scop.liasse.fiscale"].get_plancher_cotiz()[
                    "ur_med"
                ]
            self.create_contribution(
                product=product_med_id,
                type_contribution=type_cotisation_ur,
                amount=amount_med,
                journal_id=journal_ur_med_id,
            )

        # Invoice cotiz UR Ile De France
        #
        ur_idf = self.env.ref("cgscop_partner.riga_14231").id
        dom_tom_id = self.env.ref("cgscop_partner.domtom")
        if partner.ur_id.id == ur_idf and partner.region != dom_tom_id:
            product_idf_id = self.company_id.contribution_idf_id
            journal_idf_id = self.company_id.journal_ur_idf_id

            if liasse:
                amount_idf = base_cotiz.round_to_closest_multiple(
                    liasse.contribution_idf, 4
                )
            else:
                amount_idf = 840
            self.create_contribution(
                product=product_idf_id,
                type_contribution=type_cotisation_ur,
                amount=amount_idf,
                journal_id=journal_idf_id,
            )

    def create_contribution(
        self,
        product,
        type_contribution,
        amount=0,
        journal_id=False,
        type_invoice="out_invoice",
    ):
        """
        Create invoice from Contribution Base
        :param product: product_id
        :param type_contribution: type_contribution (CG, UR, Fédé)
        :param amount: contribution amount (float)
        :param journal_id: journal
        :param account_id: customer_account_id
        :param type_invoice: invoice or refund
        :return: invoice
        """
        Invoice = self.env["account.move"]
        partner = self.partner_id
        liasse = self.liasse_fiscale_id

        # Gestion du mode de paiement si un mandat existe
        mandate_ids = self.env["account.banking.mandate"].search(
            [
                ("partner_id", "=", partner.id),
                ("state", "=", "valid"),
                ("company_id", "=", self.env.company.id),
            ]
        )
        payment_mode = False
        mandate_id = False
        if mandate_ids:
            mandate_id = mandate_ids[0]
            payment_mode = self.env["account.payment.mode"].search(
                [
                    ("payment_method_id.code", "=", "sepa_direct_debit"),
                    ("company_id", "=", self.env.company.id),
                ]
            )
            self.payment_mode_id = payment_mode

        base_cotiz = self.base_cotisation_cg
        quarters = [
            base_cotiz.trimester_1,
            base_cotiz.trimester_2,
            base_cotiz.trimester_3,
            base_cotiz.trimester_4,
        ]
        type_contrib_name = (
            self.env["scop.contribution.type"].browse(type_contribution).name
        )

        for i in range(1, int(self.nb_quarter) + 1):
            cotiz_quarter = (4 - int(self.nb_quarter)) + i

            journal_id = (
                self.company_id.contribution_journal_id
                if not journal_id
                else journal_id
            )
            # Calcul prix
            price_unit = amount / 4

            Invoice.create(
                {
                    "partner_id": partner.id,
                    "liasse_fiscale_id": liasse.id,
                    "move_type": type_invoice,
                    "year": self.year,
                    "is_contribution": True,
                    "type_contribution_id": type_contribution,
                    "ref": (
                        "T"
                        + str(cotiz_quarter)
                        + " "
                        + self.year
                        + " - "
                        + type_contrib_name
                    ),
                    "journal_id": journal_id.id,
                    "state": "draft",
                    "payment_mode_id": payment_mode.id
                    if payment_mode
                    else partner.customer_payment_mode_id.id,
                    "mandate_id": mandate_id.id if mandate_id else False,
                    "partner_bank_id": mandate_id.partner_bank_id.id
                    if mandate_id
                    else False,
                    "invoice_date": self.date_cotisation,
                    "invoice_date_due": quarters[cotiz_quarter - 1],
                    "bordereau_id": self.id,
                    "amount_cg_calculated": amount,
                    "cotiz_quarter": str(cotiz_quarter),
                    "invoice_line_ids": [
                        (
                            0,
                            None,
                            {
                                "product_id": product.id,
                                "account_id": product.property_account_income_id.id,
                                "tax_ids": [(6, 0, product.taxes_id.ids)],
                                "name": product.name + " T" + str(cotiz_quarter),
                                "price_unit": price_unit,
                                "price_subtotal": price_unit,
                            },
                        ),
                    ],
                }
            )

    def update_bordereau_version(
        self,
        date_regul=None,
        comment=None,
        liasse_fiscale_new_id=None,
        special_type=None,
    ):
        """
        Crée une nouvelle version du bordereau et les appels de cotisation en brouillon
        en fonction des paramètres suivants :
        :param date_regul: Date = date des factures - date du jour par défaut
        :param comment: commentaire
        :param liasse_fiscale_new_id: id liasse fiscale de référence - liasse du
                                      bordereau initial par défaut
        :param special_type: id du type de contribution à mettre à jour si spécifié
        """
        self.ensure_one()
        bordereau_id = self
        partner_id = bordereau_id.partner_id

        if not date_regul:
            date_regul = fields.Date.today()
        if not liasse_fiscale_new_id:
            liasse_fiscale_new_id = bordereau_id.liasse_fiscale_id

        # CREATE VERSION
        bordereau_id.read(["amount_total_cotiz"])
        self.env["scop.bordereau.version"].create(
            {
                "bordereau_id": bordereau_id.id,
                "date": date_regul,
                "comment": comment,
                "version": bordereau_id.version,
                "liasse_fiscale_id_old": bordereau_id.liasse_fiscale_id.id,
                "type_assiette": bordereau_id.type_assiette,
                "montant_assiette": bordereau_id.montant_assiette,
                "amount_total_cotiz": bordereau_id.amount_total_cotiz,
            }
        )

        # CREATE REGUL
        ur_hdf = self.env.ref("cgscop_partner.riga_14232").id
        ur_med = self.env.ref("cgscop_partner.riga_14243").id
        ur_idf = self.env.ref("cgscop_partner.riga_14231").id
        type_cotisation_cg = self.env.ref("cgscop_partner.riga_14397").id
        type_cotisation_ur = self.env.ref("cgscop_partner.riga_14399").id
        type_cotisation_fede_com = self.env.ref("cgscop_partner.riga_14398").id
        type_cotisation_fede_cae = self.env.ref("cgscop_partner.cotiz_fede_cae").id
        type_cotisation_fede_indus = self.env.ref("cgscop_partner.cotiz_fede_indus").id

        # Cotisations présentes sur le bordereau
        contribs = bordereau_id.invoice_ids.read_group(
            [("id", "in", bordereau_id.invoice_ids.ids)],
            ["type_contribution_id", "amount_total_signed"],
            ["type_contribution_id"],
        )
        contrib_type = list(map(lambda c: c.get("type_contribution_id")[0], contribs))

        # Vérifie et ajoute les cotisations manquantes si nécessaire
        if type_cotisation_cg not in contrib_type:
            contribs.append(
                {
                    "type_contribution_id": (type_cotisation_cg,),
                    "amount_total_signed": 0,
                }
            )
        if (
            partner_id.ur_id.id in (ur_med, ur_hdf)
            and type_cotisation_ur not in contrib_type
        ):
            contribs.append(
                {
                    "type_contribution_id": (type_cotisation_ur,),
                    "amount_total_signed": 0,
                }
            )
        if (
            partner_id.is_federation_com
            and type_cotisation_fede_com not in contrib_type
        ):
            contribs.append(
                {
                    "type_contribution_id": (type_cotisation_fede_com,),
                    "amount_total_signed": 0,
                }
            )

        if (
            partner_id.is_federation_indus
            and type_cotisation_fede_indus not in contrib_type
        ):
            contribs.append(
                {
                    "type_contribution_id": (type_cotisation_fede_indus,),
                    "amount_total_signed": 0,
                }
            )

        if partner_id.cae and type_cotisation_fede_cae not in contrib_type:
            contribs.append(
                {
                    "type_contribution_id": (type_cotisation_fede_cae,),
                    "amount_total_signed": 0,
                }
            )

        # Variables cotisations UR
        if partner_id.ur_id.id == ur_hdf:
            amount_ur = (
                liasse_fiscale_new_id.contribution_hdf if liasse_fiscale_new_id else 40
            )
            product_ur = self.env.company.contribution_hdf_id
            journal_ur = self.env.user.company_id.journal_ur_hdf_id
        elif partner_id.ur_id.id == ur_idf:
            amount_ur = (
                liasse_fiscale_new_id.contribution_idf if liasse_fiscale_new_id else 840
            )
            product_ur = self.env.company.contribution_idf_id
            journal_ur = self.env.user.company_id.journal_ur_idf_id
        else:  # ur = ur_med
            amount_ur = liasse_fiscale_new_id.contribution_med
            product_ur = self.env.company.contribution_med_id
            journal_ur = self.env.user.company_id.journal_ur_med_id

        # Type de cotisation
        cotiz_type = {
            self.env.ref("cgscop_partner.riga_14397").id: [
                liasse_fiscale_new_id.contribution_cg if liasse_fiscale_new_id else 300,
                self.env.company.contribution_cg_id,
                self.env.company.contribution_journal_id,
                partner_id.property_account_receivable_id.id,
            ],
            self.env.ref("cgscop_partner.riga_14398").id: [
                liasse_fiscale_new_id.contribution_com
                if liasse_fiscale_new_id
                else 108,
                self.env.company.contribution_fede_com_id,
                self.env.user.company_id.journal_fede_com_id,
                partner_id.property_account_receivable_id.id,
            ],
            self.env.ref("cgscop_partner.cotiz_fede_cae").id: [
                liasse_fiscale_new_id.contribution_cae
                if liasse_fiscale_new_id
                else 300,
                self.env.company.contribution_fede_cae_id,
                self.env.user.company_id.journal_fede_cae_id,
                partner_id.property_account_receivable_id.id,
            ],
            self.env.ref("cgscop_partner.cotiz_fede_indus").id: [
                liasse_fiscale_new_id.contribution_indus
                if liasse_fiscale_new_id
                else 100,
                self.env.company.contribution_fede_indus_id,
                self.env.user.company_id.journal_fede_indus_id,
                partner_id.property_account_receivable_id.id,
            ],
            self.env.ref("cgscop_partner.riga_14399").id: [
                amount_ur,
                product_ur,
                journal_ur,
                partner_id.property_account_receivable_id.id,
            ],
        }

        quarters = [
            bordereau_id.base_cotisation_cg.trimester_1,
            bordereau_id.base_cotisation_cg.trimester_2,
            bordereau_id.base_cotisation_cg.trimester_3,
            bordereau_id.base_cotisation_cg.trimester_4,
        ]

        # Calcul des régularisations des cotisations existantes
        # pour chaque type de cotisation
        for contrib in contribs:
            type_cotiz = contrib.get("type_contribution_id")[0]
            # Montant total des appels actuels
            amount_cotiz_old = contrib.get("amount_total_signed")
            # Calcul du montant proratisé de la cotisation
            amount_cotiz = (
                self.env["scop.cotisation"].round_to_closest_multiple(
                    cotiz_type.get(type_cotiz)[0], 4
                )
                * int(bordereau_id.nb_quarter)
                / 4
            )

            # Calcul du montant total des avoirs pour une cotisation
            refund_amount_total = amount_cotiz_old - amount_cotiz

            product = cotiz_type.get(type_cotiz)[1]
            type_invoice = False

            if refund_amount_total < 0:
                type_invoice = "out_invoice"
                refund_amount_total *= -1
            elif refund_amount_total > 0:
                type_invoice = "out_refund"
            refund_amount = refund_amount_total / int(bordereau_id.nb_quarter)

            nb_quarter_int = int(bordereau_id.nb_quarter)
            if type_invoice:
                for i in range(4 - nb_quarter_int, 4):
                    if date_regul < quarters[i]:
                        date_due = quarters[i]
                    else:
                        date_due = date_regul
                    # Gestion Exception CAE pour 2022
                    if (
                        type_cotiz == type_cotisation_fede_cae
                        and self.year == "2022"
                        and i in (0, 1)
                    ):
                        pass
                    else:
                        # Gestion du type de cotisation spécifique
                        if special_type:
                            if type_cotiz == special_type:
                                self.env["account.move"].create(
                                    {
                                        "partner_id": partner_id.id,
                                        "journal_id": cotiz_type.get(type_cotiz)[2].id,
                                        "move_type": type_invoice,
                                        "invoice_date": date_regul,
                                        "date": date_regul,
                                        "state": "draft",
                                        "invoice_origin": bordereau_id.name,
                                        "ref": comment,
                                        "bordereau_id": bordereau_id.id,
                                        "is_contribution": True,
                                        "year": bordereau_id.year,
                                        "liasse_fiscale_id": bordereau_id.liasse_fiscale_id.id,
                                        "type_contribution_id": type_cotiz,
                                        "payment_mode_id": bordereau_id.payment_mode_id.id,
                                        "invoice_date_due": date_due,
                                        "cotiz_quarter": str(i + 1),
                                        "invoice_line_ids": [
                                            (
                                                0,
                                                None,
                                                {
                                                    "product_id": product.id,
                                                    "account_id": product.property_account_income_id.id,
                                                    "tax_ids": [
                                                        (6, 0, product.taxes_id.ids)
                                                    ],
                                                    "name": comment
                                                    + " - "
                                                    + str(i + 1)
                                                    + "/"
                                                    + bordereau_id.nb_quarter,
                                                    "price_unit": refund_amount,
                                                    "price_subtotal": refund_amount,
                                                },
                                            ),
                                        ],
                                    }
                                )
                        else:
                            self.env["account.move"].create(
                                {
                                    "partner_id": partner_id.id,
                                    "journal_id": cotiz_type.get(type_cotiz)[2].id,
                                    "move_type": type_invoice,
                                    "invoice_date": date_regul,
                                    "date": date_regul,
                                    "state": "draft",
                                    "invoice_origin": bordereau_id.name,
                                    "ref": comment,
                                    "bordereau_id": bordereau_id.id,
                                    "is_contribution": True,
                                    "year": bordereau_id.year,
                                    "liasse_fiscale_id": bordereau_id.liasse_fiscale_id.id,
                                    "type_contribution_id": type_cotiz,
                                    "payment_mode_id": bordereau_id.payment_mode_id.id,
                                    "invoice_date_due": date_due,
                                    "cotiz_quarter": str(i + 1),
                                    "invoice_line_ids": [
                                        (
                                            0,
                                            None,
                                            {
                                                "product_id": product.id,
                                                "account_id": product.property_account_income_id.id,
                                                "tax_ids": [
                                                    (6, 0, product.taxes_id.ids)
                                                ],
                                                "name": comment
                                                + " - "
                                                + str(i + 1)
                                                + "/"
                                                + bordereau_id.nb_quarter,
                                                "price_unit": refund_amount,
                                                "price_subtotal": refund_amount,
                                            },
                                        ),
                                    ],
                                }
                            )
        bordereau_id.update({"liasse_fiscale_id": liasse_fiscale_new_id.id})
        bordereau_id.invoice_ids.update({"liasse_fiscale_id": liasse_fiscale_new_id.id})

    def check_ongoing_version(self):
        """
        :return: le versionnage en cours
        """
        ongoing_version = self.bordereau_version_ids.filtered(
            lambda v: v.version == self.version and v.state == "new"
        )
        if len(ongoing_version) != 1:
            raise ValueError(_("Erreur sur le versionnage de ce bordereau"))
        else:
            return ongoing_version

    def get_bordereau_move_line(self):
        """
        Get move line for invoices in bordereau
        :return: dict with date as key and amount (float) as value
        """
        # TODO: voir si possible d'optimiser avec le nouveau modèle
        self.ensure_one()
        MoveLine = self.env["account.move.line"]
        # Get invoices
        inv_ids = self.invoice_ids
        # Old way : pour les bordereaux flagués régularisation (2021)
        if self.is_regul:
            inv_ids = inv_ids.filtered(lambda i: i.move_type == "out_invoice")

        schedule_plan = []
        # Bordereaux avec trimestre sur les factures
        if self.year != "2021":
            payment_schedule = inv_ids.read_group(
                domain=[("id", "in", inv_ids.ids)],
                fields=["invoice_date_due", "amount_total_signed"],
                groupby=["invoice_date_due:day"],
                orderby="invoice_date_due",
                lazy=True
            )
            schedule_plan = list(map(lambda e: {
                "date": datetime.strptime(e["__domain"][2][2], "%Y-%m-%d").strftime("%d/%m/%Y"),
                "amount": e.get("amount_total_signed")
            }, payment_schedule))

        # Old way : pour les bordereaux sans trimestre sur les factures
        else:
            move_lines = MoveLine.search(
                [
                    ("move_id", "in", inv_ids.ids),
                    ("account_id.internal_type", "=", "receivable"),
                ],
                order="date_maturity",
            )
            schedule = list(dict.fromkeys(move_lines.mapped("date_maturity")))
            schedule.sort()

            for date in schedule:
                amount = 0
                for item in move_lines:
                    if item.date_maturity == date:
                        amount += item.debit - item.credit
                schedule_plan.append(
                    {
                        "date": date.strftime("%d/%m/%Y"),
                        "amount": amount,
                    }
                )
        return schedule_plan

    def get_bordereau_move_line_with_payments(self):
        """
        Get payment dates for invoices in bordereau
        :return: dict with date as key and amount (float) as value
        """
        Move = self.env["account.move"]
        for bordereau in self:
            payment_ids = bordereau.invoice_ids.move_line_payment_ids.filtered(
                lambda r: r.move_id.move_type == "entry"
            ).sorted(lambda p: p.date)

            payments = []
            for key, group in itertools.groupby(payment_ids, lambda p: p.date):
                pay_ids = list(group)
                credit = sum(list(map(lambda p: p.credit, pay_ids)))
                debit = sum(list(map(lambda p: p.debit, pay_ids)))
                payments.append(
                    {
                        "date": key,
                        "credit": credit,
                        "debit": debit,
                    }
                )

            schedule_plan = Move.read_group(
                [
                    ("bordereau_id", "=", bordereau.id),
                ],
                fields=[
                    "invoice_date_due",
                    "amount_total_signed",
                    "amount_residual_signed",
                ],
                groupby=["invoice_date_due:day"],
                orderby="invoice_date_due",
                lazy=False,
            )
            for contribution in schedule_plan:
                contribution["date_maturity"] = contribution["__domain"][2][2]

            total_amount = sum(
                list(map(lambda l: l.get("amount_total_signed"), schedule_plan))
            )
            total_residual = sum(
                list(map(lambda l: l.get("amount_residual_signed"), schedule_plan))
            )
            total_paid = total_amount - total_residual
            return {
                "plan": schedule_plan,
                "payments": payments,
                "total_amount": total_amount,
                "total_paid": total_paid,
                "total_residual": total_residual,
            }

    def get_contribution_type(self):
        """
        Get contribution by type
        :return: dict type_contribution and amount
        """
        domain = [("id", "in", self.invoice_ids.ids)]
        if self.is_regul:
            domain.append(("move_type", "=", "out_invoice"))
        return self.invoice_ids.read_group(
            domain,
            ["type_contribution_id", "amount_total_signed"],
            ["type_contribution_id"],
        )

    def get_contribution_type_refund(self):
        """
        Get contribution by type for refund
        :return: dict type_contribution and amount
        """
        return self.invoice_ids.read_group(
            [("id", "in", self.invoice_ids.ids), ("move_type", "=", "out_refund")],
            ["type_contribution_id", "amount_total_signed"],
            ["type_contribution_id"],
        )

    # Email
    def get_recipients(self):
        recipients = ",".join(map(lambda x: str(x), self.recipient_ids.ids))
        return recipients
